/*
 data-ooc-isec-local.c

 Parallel version of simple irregular loop (see `seq.c' file for details).
 Index arrays are out-of-core 
 Data arrays are out of core so they are to be read from disk using MPI IO
 routines.
 Data array of each node is divided into in-core-sections (i-sections) which
 are useful when there is not enough memory to place data from the disk 
 inone slice.
 For simplicity of the code all arrays are assumed to be evenly divisible among
 all nodes.
 Each node opens its locall files possibly on local filesystems esp. on clusters
 with local disks on each node.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>


#include <lip.h>
#include <debug.h>
/* size of data array can be evenly divided by number 1...10 (which are allowed
 * total numbers of nodes) */

#define N (2 * 3 * 2 * 5 * 7 * 2 * 9 )
/*#define N (2 * 3 * 2)*/
#define L (2 * N)
#define ISEC 6 /*6 number of i-sections*/
/* ceiling division from HPF 2.0 */
#define CD(j,k) (((j)-1)/(k)+1)
#define FILE_DIR "."



int
main(int argc, char **argv)
{
  double *x, *y, *x_l1, *x_l2;
  int *perm, *perm_l, *tperm_l;
  int n = N,   l = L;
  int n_l,     l_l;    /* numbers of elements on single node (local elements) */
  int n_start, l_start; /* first index on this node */
  int isec_n_l, isec_l_l;
  int isec_cnt;
  int ret;
  char * file_name;
  struct timeval tv1, tv2;

  int i, s, p;
  double sum_y_l, sum_y, sum_x_l, sum_x, z;
  double *sum_y_isec; /*table to store local sums of each i-section*/

  int rank, size; /* LIP's environment parameters */
  
  MPI_Info info; /* needed by MPI IO */
  MPI_File file1, file2; /* we need two files : 1 for input, 2 for output data, 
                            to handle 2 irregular loops at one time */
  MPI_File ifile;        /* file for index array */
  MPI_Offset offset;
  MPI_Status status;


  LIP_Datamap datamap;
  MPI_Datatype filetype, memtype;
  LIP_Schedule schedule;
  LIP_IOBufmap bufm;	

  MPI_Init( &argc, &argv );
  LIP_Setup( NULL, 0 ); /* all processes go to LIP */
  
  MPI_Info_create(&info);

  filetype = MPI_DATATYPE_NULL;
  memtype = MPI_DATATYPE_NULL;
  LIP_Rank( &rank );
  LIP_Size( &size );

  l_l = CD( l, size ); /* number of `x' elements on single node disk*/
  n_l = CD( n, size ); /* number of `y' & `perm' elements on single node disk*/
  
  isec_l_l = l_l/ISEC; /* sizes of arrays in mamory */
  isec_n_l = n_l/ISEC; 

  l_start = l_l * rank;
  n_start = n_l * rank;

  /*** Allocate  memory ***/

  file_name = malloc(100);

  y    = malloc( isec_n_l * (sizeof *y) );

  x_l1    = malloc( (isec_n_l + isec_n_l) * (sizeof *x) ); /* additional space for non-local
                                            data (which may be of size `isec_n_l') */
  x_l2    = malloc( (isec_n_l + isec_n_l) * (sizeof *x) ); /* additional space for non-local
                                            data (which may be of size `isec_n_l') */
  sum_y_isec = malloc (ISEC * (sizeof *sum_y_isec));

  perm   = malloc( isec_n_l * (sizeof *perm) ); /* index array that is to be divided into i-sections; 
                                              in inex out-of-core problems it should be located on a disk */
  perm_l = malloc( isec_n_l * (sizeof *perm) );      /* memory to contain one i-section of indices */
  tperm_l = malloc( isec_n_l * (sizeof *perm) );
  
 /*** Fill data files ***/

  sprintf(file_name, "%s/data.file%i", FILE_DIR, rank);
  MPI_File_open(MPI_COMM_WORLD, file_name, MPI_MODE_RDWR | 
                MPI_MODE_CREATE, MPI_INFO_NULL, &file1);
  MPI_File_set_view(file1,l_start*8,MPI_DOUBLE, MPI_DOUBLE, "native", 
MPI_INFO_NULL);

  sprintf(file_name, "%s/data-out.file%i", FILE_DIR, rank);
  MPI_File_open(MPI_COMM_WORLD, file_name, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &file2);
  MPI_File_set_view(file2,l_start*8,MPI_DOUBLE, MPI_DOUBLE, "native", 
MPI_INFO_NULL);

  sprintf(file_name, "%s/index.file%i", FILE_DIR, rank);
  MPI_File_open(MPI_COMM_WORLD, file_name, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &ifile);
  MPI_File_set_view(ifile,n_start*4,MPI_INT, MPI_INT, "native", 
MPI_INFO_NULL);



  /*** Fill data  array ***/

  z = .1313;
  for (i = 0 ; i < l_start ; i++) /* skip values for previous nodes */
    z = 4.0 * z * (1.0 - z);

  for(isec_cnt=0; isec_cnt<ISEC; isec_cnt++)
  {
       for (i = 0 ; i < isec_l_l ; i++) /* generate values for this i-section */
       {
       x_l1[i] = z - 0.5;
  
       z = 4.0 * z * (1.0 - z);
         /*x[i] = l_start+i;*/
       }
       offset = isec_cnt * isec_l_l; /* Each node writes its part of a file */
       MPI_File_write_at(file1,offset,x_l1,isec_l_l,MPI_DOUBLE,&status); 
       MPI_File_write_at(file2,offset,x_l1,isec_l_l,MPI_DOUBLE,&status);    
  }
 
  /*** Fill index array ***/

  /* skip values for previous nodes */
  for (p = .1313 * l, s = -l, i = 0 ; i < n_start ; i++, s += 2)
    p = (p + s + l) % l;
  for(isec_cnt=0; isec_cnt<ISEC; isec_cnt++)
  {
       offset = isec_cnt * isec_n_l;
       for (i = 0 ; i < isec_n_l ; i++, s += 2) /* generate values for this node */
       {
         perm[i] = p;
         p = (p + s + l) % l;
       }
       MPI_File_write_at(ifile,offset,perm,isec_n_l,MPI_INT,&status); 
  }

  /**** Start timer ****/

   gettimeofday(&tv1, 0);

  

  /*** Inspector  phase ***/
  /* create IO bufmap which maps local data buffer on the data array on disk */ 
  LIP_IOBufmap_create(isec_n_l, &bufm);
  
  /* create data map which describes how data array `x' is partitioned on a disk */
  LIP_Datamap_create( LIP_DATAMAP_BLOCK, l, &l_l, LIP_INDEXTYPE_INT, 1, 0,
      &datamap );
 
   LIP_Schedule_create( l_l, &schedule ); /* l_l - size of array on disk */
  /*** for each i-section ***/
  for(isec_cnt=0;isec_cnt<ISEC; isec_cnt++)
  {

     offset = isec_cnt * isec_n_l;
     MPI_File_read_at(ifile,offset,perm,isec_n_l,MPI_INT,&status); 
    ret = LIP_Localize( datamap, perm, LIP_INDEXTYPE_INT, perm_l, LIP_INDEXTYPE_INT,
      isec_n_l, 0, &schedule, LIP_INDEXINFO_NULL ); /* as index array we choose only one i-section
                                                per one step (perm+(isec_n_l*isec_cnt)) */
   if (ret!=LIP_SUCCESS) printf ("Fatal error!\n");

    
    /* schedule has to be committed before its use in communication functions */
    LIP_Schedule_commit( &schedule );

    if(LIP_SUCCESS!=LIP_OOC_localize(perm_l, LIP_INDEXTYPE_INT, tperm_l, LIP_INDEXTYPE_INT, 
    		isec_n_l, 0, x_l1, MPI_DOUBLE, &bufm, &schedule)) printf("BLAD!!!\n");

   /* After OOC_Localize we know 
   which data are not used in next step, so they can be written to the disk*/ 
   ret=LIP_IOBufmap_get_datatype(bufm,MPI_DOUBLE,&memtype,&filetype,LIP_DATA_OLD);
   if(ret){
    printf("Tu blad !!!! ");
   }
    /* set file view according to calculated mapping */  
    /* Displacement (in bytes) is calculated according to block data mapping*/

    if (MPI_DATATYPE_NULL!=memtype) {
      
      MPI_File_set_view(file2,8*l_start,MPI_DOUBLE,filetype,"native",info);
      MPI_File_write_at(file2,0, x_l2, 1, memtype, &status);
     
      MPI_Type_free(&memtype);
      MPI_Type_free(&filetype);
    }
    else{
      MPI_File_set_view(file2,0,MPI_DOUBLE,MPI_DOUBLE,"native",info);
      MPI_File_write_at(file2,0, x_l2, 0, MPI_DOUBLE, &status);
/*         printf("puste write %d %d\n ",rank,isec_cnt);*/
    }
   
    /* After updating arrays on disk we can now read new data to a memory */
    LIP_IOBufmap_get_datatype(bufm,MPI_DOUBLE,&memtype,&filetype,LIP_DATA_NEW);
    if (MPI_DATATYPE_NULL!=memtype) {
    
      MPI_File_set_view(file1,8*l_start,MPI_DOUBLE,filetype,"native",info);
      MPI_File_set_view(file2,8*l_start,MPI_DOUBLE,filetype,"native",info);
    /* read from files */
      MPI_File_read_at(file1,0, x_l1, 1, memtype, &status);
      MPI_File_read_at(file2,0, x_l2, 1, memtype, &status);
      
/*        MPI_File_set_view(file1,0,MPI_DOUBLE, MPI_DOUBLE, "native", MPI_INFO_NULL);*/
/*        MPI_File_set_view(file1,0,MPI_DOUBLE, MPI_DOUBLE, "native", MPI_INFO_NULL);*/
       MPI_Type_free(&memtype);
      MPI_Type_free(&filetype);
    }
    else{
      MPI_File_set_view(file1,0,MPI_DOUBLE,MPI_DOUBLE,"native",info);
      MPI_File_set_view(file2,0,MPI_DOUBLE,MPI_DOUBLE,"native",info);
      MPI_File_read_at(file1,0, x_l1, 0, MPI_DOUBLE, &status);
      MPI_File_read_at(file2,0, x_l2, 0, MPI_DOUBLE, &status);
    
    }

    /*** Executor   phase ***/

    /* gather non-local data from other nodes */
    LIP_Gather( x_l1, x_l1 + isec_n_l, MPI_DOUBLE, schedule );

    /*** irregular loops ***/

    for (i = 0 ; i < isec_n_l ; i++)
      y[i] = -x_l1[ tperm_l[i] ]; 

    /* clear the ghost area */
    for (i = 0 ; i < isec_n_l ; i++)
      x_l2[ i + isec_n_l ] = 0.0; 
   
    for (i = 0 ; i < isec_n_l ; i++)
       x_l2[tperm_l[i] ] += y[i];
       
    /* scatter non-local data to other nodes */
    LIP_Scatter( x_l2 + isec_n_l, x_l2, MPI_DOUBLE, schedule, LIP_OP_SUM );

    for (sum_y_l = 0.0, i = 0 ; i < isec_n_l ; i++)
      sum_y_l += y[i];
    sum_y_isec[isec_cnt] = sum_y_l;


  } /*end for each i-section*/

  LIP_Schedule_free( &schedule );
  /* save all data to the disk */
  ret=LIP_IOBufmap_get_datatype(bufm,MPI_DOUBLE,&memtype,&filetype,LIP_DATA_ALL);
  if (MPI_DATATYPE_NULL!=memtype) {

  MPI_File_set_view(file2,8*l_start,MPI_DOUBLE,filetype,"native",info);
  MPI_File_write_at(file2,0, x_l2, 1, memtype, &status);
      MPI_Type_free(&memtype);
      MPI_Type_free(&filetype);
  }
  else{
    MPI_File_set_view(file2,0,MPI_DOUBLE,MPI_DOUBLE,"native",info);
    MPI_File_write_at(file2,0, x_l2, 0,MPI_DOUBLE, &status);
  
  }
  /* read modified x array to calculate final sum */  
  MPI_File_set_view(file2,8*l_start,MPI_DOUBLE, MPI_DOUBLE, "native", 
MPI_INFO_NULL);

  sum_x_l = 0.0;
  for(isec_cnt=0; isec_cnt<ISEC; isec_cnt++)
  {       
       offset = isec_cnt * isec_l_l; /* Each node writes its part of a file */
       MPI_File_read_at(file2,offset,x_l2,isec_l_l,MPI_DOUBLE,&status);    
       for (i = 0 ; i < isec_l_l ; i++) /* generate values for this i-section */
       {
           sum_x_l += x_l2[i];
       }
  }


  LIP_Datamap_free( &datamap, 0 );


  if (0 == rank && size > 1)
    printf( "%d computing nodes. "
      "Each has %d `y' elements and %d `x' elements.\n", size, n_l, l_l );
      

  for (sum_y_l = 0.0, i = 0 ; i < ISEC ; i++)
    sum_y_l += sum_y_isec[i];

  MPI_Reduce( &sum_y_l, &sum_y, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `y' is %g\n", n, sum_y );

 
  MPI_Reduce( &sum_x_l, &sum_x, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

  if (0 == rank)
    printf( "The sum of %d elements of `x' is %g\n", l, sum_x );

  gettimeofday(&tv2, 0);
  printf("Rank %i - computation time in microseconds %d\n",
          rank, 1000000*(tv2.tv_sec - tv1.tv_sec) + tv2.tv_usec - tv1.tv_usec);




  MPI_File_close(&file1); /* Close a file before calling MPI_Finalize! */
  MPI_File_close(&file2); /* Close a file before calling MPI_Finalize! */
  MPI_File_close(&ifile); /* Close a file before calling MPI_Finalize! */
 
  LIP_Exit();
  MPI_Finalize();


  free( perm_l );
  free( perm );
  free( y );
  free( x_l1 );
  free( x_l2 );

  return 0;
} /* main */

