/*
 seq.c

 Sequential version of simple irregular program.

 Irregular loops in the program are:
 for (i = 0 ; i < n ; i++)
   y[i] = -x[ perm[i] ];

 for (i = 0 ; i < n ; i++)
   x[ perm[i] += y[i];

 The outcome of the loops can be easily checked by summing up the entries from
 `x' and `y' arrays. The sum should be the same for sequential and parallel
 versions of the program.

 Arrays used (x,y - data arrays; perm - index array) can be sketched like this:

 y[]  perm[]   x[]
 +-+ -- +-+ -- +-+
 | |    | |    | |
 | |    | |    | |
 | |    | |    | |
 +-+ -- +-+    | |
           \   | |
	    \  | |
	     \ | |
	      \| |
	       +-+

 */

#include <stdio.h>
#include <stdlib.h>
#include "generate.h"

#define N (2 * 3 * 2 * 5 * 7 * 2 * 9)
#define L (2 * N)

/* `x' is evaluated twice ! */
#define ABS(x) ((x)<0?-(x):(x))
#define EPSILON 1e-5



int
main(void)
{
  double *x, *y;
  int *perm;
  int n = N, l = L;

  int i, s, p;
  double sum_x, sum_y, z;

  x    = malloc( l * (sizeof *x) );
  y    = malloc( n * (sizeof *y) );
  perm = malloc( n * (sizeof *perm) );

  generate_data(x, perm, l, 0, l, 0, n);


  /* irregular loop */
  for (i = 0 ; i < n ; i++)
    y[i] = -x[ perm[i] ];

  for (i = 0 ; i < n ; i++)
    x[ perm[i] ] += y[i];

  for (sum_y = 0.0, i = 0 ; i < n ; i++)
  {
    sum_y += y[i];
  }

  printf( "The sum of %d elements of `y' is %g\n", n, sum_y );

  for (sum_x = 0.0, i = 0 ; i < l ; i++)
  {
    sum_x += x[i];
  }

  printf( "The sum of %d elements of `x' is %g\n", l, sum_x );

  free( perm );
  free( y );
  free( x );

  return 0;
} /* main */
