#ifndef LIPIMPL_H
#define LIPIMPL_H

#include <lip.h>

typedef struct {
  int rank;
  int size;
  MPI_Comm comm;
  int initialized;
} LIP_Env;

extern LIP_Env LIP_env;

typedef struct {
  int *esizes;          /* export sizes (size = number of nodes) */
  int *edispls;         /* export displacements (size = number of nodes)
                         * This array could be easly generated from 'esizes',
                         * but in order to increase speed it is generated only
                         * once upon creation of schedule. */
  int *elocations;      /* export locations
                         * (size = edispls[last] + esizes[last]; where 'last'
                         * is the last index in arrays 'edispls' and 'esizes')
                        */
  int *isizes;          /* import sizes (size = number of nodes) */
  int *idispls;         /* import displacements (size = number of nodes)
                         * This array could be easly generated from 'isizes'.
                         * See discussion on 'edispls'.*/
  int data_size;	/* size of local data array */

  int committed;	/* says if schedule is committed */
  void *user_buf;	/* communication buffer given by user */
} LIP_ISchedule;
typedef struct {
    int index;
    unsigned int h_index;
} LIP_bucket_element;
typedef struct {
    int g_index;
    int owner;
} LIP_ttable_change;

typedef struct {
  int type;
  int global_size;/*size of all data (sum of all nodes)*/
  int local_size;/*size of the single node*/
  int node_size;
  int ttable_ic_size;
  int ttable_ic_real_size; /*ic size+ghost area*/
  int ttable_ooc_size;
  int* procs;
  int* local_adr;
  MPI_File procs_file;
  MPI_File local_adr_file;
  char * procs_fname;
  char * local_adr_fname;
  LIP_IOBufmap tt_io_bmap;
  int flag;/*LOCAL or NONLOCAL*/
} LIP_IDatamap;

#define LIP_DEREF_ALL 3

LIP_IDatamap *LIP_Datamap_new(LIP_Datamap *datamap);
LIP_IDatamap *LIP_Datamap_convert(LIP_Datamap m);
int LIP_Datamap_array_init();
int LIP_Datamap_array_free();


int LIP_Dereference(LIP_Datamap datamap, void *iglobal, LIP_Indextype igtype,
  void *ilocal, LIP_Indextype iltype, int *procs, int count, int deref_flag,
  int* indcs_on_proc);


LIP_ISchedule *LIP_Schedule_convert(LIP_Schedule s);
int LIP_Schedule_array_init();
int LIP_Schedule_array_free();
/* OOC */

typedef struct {            /*MM & TB*/
  int size;                 /* memory buffer size */
  int * oldhash_val; 
  int * oldhash_key; 
  int * newhash_val; 
  int * newhash_key; 
  int htab_size;            /* hash table size */
} LIP_IIOBufmap;


LIP_IIOBufmap *LIP_IOBufmap_convert(LIP_IOBufmap m); /*MM*/
int LIP_IOBufmap_array_init();
int LIP_IOBufmap_array_free();

/*sets name of procedure to report errors*/
int LIP_set_proc_name();
#endif /* ifndef LIPIMPL_H */
