/*LIP library 
 errors.c - file that contains errors returned by LIP functions
 created by Katarzyna Zajac 
*/
#include <lipimpl.h>
#include <string.h>
/*Name of the procedure - it is used to report errors*/

int LIP_Error_mesg(int error_code,char* error_str,int* length){
 const char* new_mesg;
 error_str[0]=0;
 switch(error_code){
    case LIP_SUCCESS:
	strcpy(error_str,"no error");
	break;
    case LIP_OUT_OF_MEMORY:
	strcpy(error_str,"not enough memory");
	break;
    case LIP_NOT_SUPPORTED:
	strcpy(error_str,"not supported by lip");
	break;
    case LIP_INVALID_ARGUMENT:
	strcpy(error_str,"invalid argument ");
	break;
    case LIP_TOO_MANY_HANDLES:
	strcpy(error_str,"no space for creating new handle");
	break;
    case LIP_NOT_COMMITED:
	strcpy(error_str,"type not commited:");
	break;
    	
    case LIP_ERROR:
	strcpy(error_str,"error");
	break;
    default:
	strcpy(error_str,"unknown error");
	break;
    
 }
 *length=strlen(error_str);
 return LIP_SUCCESS;
}

/*The procedure prints error_mesg and quits*/
int LIP_Error_quit(char*proc_name,char* error_mesg,int error_code){
 int length;
 char error_str[LIP_MAX_ERROR_STR_LENGTH];
 LIP_Error_mesg(error_code,error_str,&length);
 if(NULL!=proc_name){
 if (NULL!=error_mesg){
   fprintf(stderr,"rank %d %s: %s %s\n",LIP_env.rank,proc_name,error_str,error_mesg);
 }
 else{
  fprintf(stderr,"rank %d %s: %s\n",LIP_env.rank,proc_name,error_str);
 }
 }
 else{
 if (NULL!=error_mesg){
   fprintf(stderr,"rank %d: %s %s\n",LIP_env.rank,error_str,error_mesg);
 }
 else{
  fprintf(stderr,"rank %d: %s\n",LIP_env.rank,error_str);
 }
 }
/* LIP_Exit(); */
 MPI_Abort(LIP_env.comm,error_code);
 abort();
 return LIP_ERROR;
}
/*Prints error_mesg and returns to the program*/
int LIP_Error_ret(char*proc_name,char* error_mesg,int error_code){
 int length;
 char error_str[LIP_MAX_ERROR_STR_LENGTH];
 
 LIP_Error_mesg(error_code,error_str,&length);
 if(NULL!=proc_name){
  if (NULL!=error_mesg){
   fprintf(stderr,"rank %d %s: %s %s\n",LIP_env.rank,proc_name,error_str,error_mesg);
  }
  else{
   fprintf(stderr,"rank %d %s: %s\n",LIP_env.rank,proc_name,error_str);
  }
 }
 else{
  if (NULL!=error_mesg){
   fprintf(stderr,"rank %d: %s %s\n",LIP_env.rank ,error_str,error_mesg);
  }
  else{
   fprintf(stderr,"rank %d: %s\n",LIP_env.rank,error_str);
  }
 }
 return LIP_SUCCESS;
}
/*If ptr is NULL the procedure prints error_mesg and quits
 in the opposite case it returns LIP_SUCCESS*/
int LIP_Assert(void* ptr,char*proc_name,char* error_mesg,int error_code){
 if(ptr){
   return LIP_SUCCESS;
 }
 else{
  LIP_Error_quit(proc_name,error_mesg,error_code);
  return LIP_ERROR;
 }
}

