/* 
LIP library
File liptools.c
created by Maciej Malawski19.11.1999 *
moved from: localize.c & adapted
*/

#include <liptools.h>
#include <lip.h>
#include <lipimpl.h>

/*
 floor( log( n ) / log( 2 ) )
 or: floor( log2( n ) )
 */
int log2(unsigned int n)
{
  int log;

  for (log = 0 ; n > 1 ; n >>= 1, log++)
    ;

  return log;
}

/* Return smallest prime greater than 'n' which is a Radke number. To determine
 * whether number is prime simple division method is used (subsequent division
 * are performed by numbers: 5 and 6*i+1, 6*i+5; i=1,2,3,...,sqrt(n)/6). This
 * may lead to very long computation time when 'int' has more than 32 bits
 * (e.g. 64) and 'n' has big value. */
unsigned int prime_round_up(unsigned int n)
{
# define CNDT_COUNT 20
  unsigned int candidates[CNDT_COUNT + 1];
  int i;
  unsigned int div, incr, max_div;
  int cndt_gone = CNDT_COUNT;
  int break_loop;
  if (n < 11)
    return 11;

  if (1 != (i = n % 6))
    n += 7 - i;

  if (((n - 3) % 4)) /* if n-3 is not divisible by 4(it can't be Radke number)*/
    n += 6; /* now n-3 is divisible by 4 (it can be Radke number) */

  while (cndt_gone == CNDT_COUNT)
  {
    cndt_gone = 0;

    /* fill array of candidates that could be Radke numbers */
    for (i = CNDT_COUNT ; i >= 0 ; i--)
      candidates[i] = n + 6 * (i - (i & 1)) + i % 2 * 4;

    n = candidates[CNDT_COUNT];

    max_div = 1 << (log2( candidates[CNDT_COUNT - 1] ) + 2) / 2;

    break_loop = 0;
    for (incr = 2, div = 5 ; div <= max_div && cndt_gone < CNDT_COUNT ;
      div += incr, incr = incr == 2 ? 4 : 2)
    {
      for (i = CNDT_COUNT - 1 ; i >= 0 ; i--)
      {
        if (candidates[i] != 0 && candidates[i] % div == 0)
        {
          candidates[i] = 0;
          cndt_gone++;
	  if (cndt_gone == CNDT_COUNT)
          { 
	     break_loop = 1;
             break;
             /* goto breakboth;*/ /* break both loops */
          }
        }
      }
      if (break_loop) break;
    }
/*    breakboth:*/
  }

  for (i = 0 ; i < CNDT_COUNT ; i++)
    if (candidates[i])
      return candidates[i];

  return 0;

# undef CNDT_COUNT
}


/*
Function computes index in hashtable where data indexed by key will be stored 
and returns it in *hash_index. 
*empty = 1 means the key is being inserted into hashtable for the first time 
 and *hash_index points to its new place .
*empty = 0 means it has been already inserted into hashtable 
 and *hash_index points to that location.
*/

int hash_query (int key, int * hash_index, int * empty, int * hash_vals, 
                int * hash_keys, int htab_size)
{
int h_ind;
int j;
  /* compute hash index; for best results `htab_size' should be prime
            * number */
  h_ind = key % htab_size;
  if (HASH_EMPTY == hash_vals[h_ind]) 
    *empty = 1;
  else if (hash_keys[h_ind] == key)
    *empty = 0;
  else
  {
    j =- htab_size;
    /* look for empty position */
    do
    {
      j += 2;
      h_ind += ABS(j);
      if (h_ind >= htab_size)
        h_ind -= htab_size;
    } while (hash_keys[h_ind] != key &&
             hash_vals[h_ind] != HASH_EMPTY && j < htab_size);
    /* What if impossible happens: hash table is full? */
    if (j >= htab_size){
     LIP_Error_ret(NULL,"hash table is full",LIP_ERROR);
     return LIP_ERROR;     /*MM*/

    } 
    if (HASH_EMPTY == hash_vals[h_ind])
      *empty = 1;        
    else if (hash_keys[h_ind] == key) 		    
	    *empty = 0;      
  } /* end of hash query */ 
  *hash_index = h_ind;
  return LIP_SUCCESS;
}  	
int LIP_copy_files (MPI_File data_file,
                   MPI_File data_out_file,
                   MPI_Offset offset,int count,
                   int page_size,MPI_Datatype datatype)
{
                        
    MPI_Info info;
    MPI_Status status;
    int i,j,count_rest;
    void * buf;
    int rank,size;
                        
    int rest_page,rest,datatype_size;
    char*pname="Copy_files";
    MPI_Info_create(&info);
     
    LIP_Rank(&rank);
    LIP_Size(&size);
    
    MPI_Type_size(datatype,&datatype_size);
 
    buf=(void*)malloc(page_size*datatype_size);

    LIP_Assert(buf,pname,"not enough memory for allocating data for moving",
                   LIP_OUT_OF_MEMORY);

 

    MPI_File_set_view(data_file, offset*datatype_size,
datatype,datatype,"native", info);

    MPI_File_set_view(data_out_file, 0, datatype,datatype,
"native",info);
    rest=count;
    while(rest>=page_size){

         MPI_File_read(data_file, buf,page_size, datatype, &status);

            MPI_File_write(data_out_file,
                        buf,page_size, datatype, &status);
            rest-=page_size;
    }
    if(rest>0){
            MPI_File_read(data_file, buf,rest, datatype, &status);
            MPI_File_write(data_out_file,
                        buf,rest, datatype, &status);



    }

    free(buf);
    return LIP_SUCCESS;

}

