/*
 LIP library
 remap.c - contains functions for data remapping in core
 created by Katarzyna Zajac
 */

#include <stdlib.h>
#include <string.h>

#include <lipimpl.h>
#include <liptools.h>
#include <debug.h>



int
LIP_remap_in_core(LIP_Datamap olddatamap,LIP_Datamap newdatamap,
  LIP_Schedule *schedule)
{
  int i, j,count;

  int data_size;

  LIP_IDatamap *iomap,*inmap;
 
  int *left_disp,*left_disp_next,*procs,*local_ind;

  MPI_Status status;

  LIP_ISchedule *s;
  
  char * pname="LIP_remap_in_core";
  
  if (1 == LIP_env.size)
  {
    *schedule = LIP_SCHEDULE_NULL;

    return LIP_SUCCESS;
  }
  
  iomap=LIP_Datamap_convert(olddatamap);
  inmap=LIP_Datamap_convert(newdatamap);
  
  if(NULL==iomap || NULL==inmap){
   LIP_Error_quit(pname,"one of datamaps is invalid",LIP_INVALID_ARGUMENT);
  }
  if(LIP_DATAMAP_BLOCK!=iomap->type){
    LIP_Error_quit(pname,"remapping can only be done when old datamap is BLOCK!",LIP_NOT_SUPPORTED);
  }
  if(NULL==inmap->procs){
    LIP_Error_quit(pname,"no processors table in new datamap",LIP_INVALID_ARGUMENT); 
  }
  /*get information that is already in a new datamap*/
  /*get new owners of the data*/ 
  count=inmap->ttable_ic_size;
  
  procs=inmap->procs;
  
  local_ind=inmap->local_adr=malloc(sizeof(*local_ind)* inmap->ttable_ic_real_size);
  
  LIP_Assert(local_ind,pname,"too big size of translation table - try to decrease GHOST_AREA_RATE",LIP_OUT_OF_MEMORY);

  s = LIP_Schedule_convert( *schedule );
  
  LIP_Assert(s,pname,"schedule",LIP_INVALID_ARGUMENT);
  /*allocating space for schedule*/
  data_size = s->data_size;


  s->esizes = malloc( (sizeof *s->isizes) * LIP_env.size );
  LIP_Assert(s->esizes,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
  
  s->edispls = malloc( (sizeof *s->idispls) * LIP_env.size );
  LIP_Assert(s->edispls,pname,"while filling schedule",LIP_OUT_OF_MEMORY);
 
  
  for(i=0;i<LIP_env.size;i++){
   s->esizes[i]=0;
   s->edispls[i]=0;
  }
  /*counting local indices*/
  left_disp= malloc (sizeof(*left_disp)* LIP_env.size);
  LIP_Assert(left_disp,pname,NULL,LIP_OUT_OF_MEMORY);

  left_disp_next= malloc (sizeof(*left_disp)* LIP_env.size);
  LIP_Assert(left_disp_next,pname,NULL,LIP_OUT_OF_MEMORY);
  
  for(i=0;i<count;i++){
   j=s->esizes[procs[i]]++;
  }

  MPI_Scan(s->esizes,left_disp,LIP_env.size,MPI_INT,MPI_SUM,LIP_env.comm);  
  for(i=0;i<LIP_env.size;i++){
     left_disp[i]-=s->esizes[i];
  }    
  
  for (s->edispls[0] = 0, i = 1 ; i < LIP_env.size ; i++)
  {
    s->edispls[i] = s->edispls[i - 1] + s->esizes[i - 1];
    s->esizes[i - 1] = 0; 
  }
  s->esizes[LIP_env.size - 1] = 0;

  s->elocations = malloc( (sizeof *s->elocations) *(count ));
   
  LIP_Assert(s->elocations,pname,"too many data for exporting at once",LIP_OUT_OF_MEMORY);
  

  for (i = 0 ; i < count ; i++)
  {
	j = s->edispls[ procs[i] ] + s->esizes[ procs[i] ];
    
	/* location on foreign node */
        s->elocations[j] = i;
        local_ind[i]=s->esizes[procs[i]]+left_disp[procs[i]];
        s->esizes[ procs[i] ]++; /* position for next index */

  }

  s->isizes  = malloc( (sizeof *s->isizes) * LIP_env.size );
  LIP_Assert(s->isizes,pname,"while filling schedule",LIP_OUT_OF_MEMORY);

  s->idispls = malloc( (sizeof *s->idispls) * LIP_env.size );
  LIP_Assert(s->isizes,pname,"while filling schedule",LIP_OUT_OF_MEMORY);

  /* `isizes' is copied from other nodes' `esizes' array, therfore information
   * stored in the former array is reduntant but with this redundancy
   * no communication is required whenever `isizes' is to be used */
  MPI_Alltoall( s->esizes, 1, MPI_INT, s->isizes, 1, MPI_INT, LIP_env.comm );

  /* create array of export displacements */
  for (s->idispls[0] = 0, i = 1 ; i < LIP_env.size ; i++)
    s->idispls[i] = s->idispls[i - 1] + s->isizes[i - 1];

  return LIP_SUCCESS;
} /* LIP_Remap */

