
#include <lip.h>
#include <mpi.h>
#include <lipimpl.h>
#include <schedule_ooc.h>


static LIP_ISchedule_ooc *ooc_schedule_array[MAX_SCHEDULE];

static int ooc_schedule_index = 0;

int LIP_OOC_Schedule_array_init(){
 int i;
 for(i=0;i<MAX_SCHEDULE;i++){
  ooc_schedule_array[i]=NULL;
 } 
 return LIP_SUCCESS;
}

int LIP_OOC_Schedule_array_free(){
 int i;
 LIP_Schedule_ooc sch;
 for(i=0;i<MAX_SCHEDULE;i++){
  if(ooc_schedule_array[i]!=NULL){
   sch=i+1;
   LIP_OOC_Schedule_free(&sch);
  }
 } 
 return LIP_SUCCESS;
}

int LIP_OOC_Schedule_create (LIP_Schedule_ooc * so)
{
     MPI_Info info;
     LIP_ISchedule_ooc * s;
     int i;
     int found;
     char*pname="LIP_Schedule_create";
     MPI_Info_create(&info);
     if(NULL!=ooc_schedule_array[ooc_schedule_index])
     {
        found=0;
        for(i=0;i<MAX_SCHEDULE;i++)
        {
              if (NULL==ooc_schedule_array[i])
              {
               ooc_schedule_index=i;
               found=1;
               break;
              }
        }
        if(!found)
        {
          LIP_Error_ret(pname,"MAX_SCHEDULE is too small",LIP_TOO_MANY_HANDLES);
          return LIP_TOO_MANY_HANDLES;
        }
     }
  
     /*schedule creation*/
     ooc_schedule_array[ooc_schedule_index] = s = (LIP_ISchedule_ooc *) malloc( sizeof *s );
  
     if (NULL == s)
     {
          *so=LIP_SCHEDULE_NULL;
          LIP_Error_ret(pname,"for creating ooc schedule",LIP_OUT_OF_MEMORY);
          return LIP_OUT_OF_MEMORY;
     }
     s->fname = (char *) malloc(20);
     printf("Temp dir: %s\n", getenv("LIP_TEMP"));
     s->fname = (char *) tempnam((char*)getenv("LIP_TEMP"), ".lip");
     MPI_File_open(MPI_COMM_SELF, s->fname, MPI_MODE_RDWR|MPI_MODE_CREATE,
                   info, &s->file);
     MPI_File_set_view(s->file, 0, MPI_INT, MPI_INT, "native", info);
     *so = ++ooc_schedule_index;
     ooc_schedule_index=ooc_schedule_index % MAX_SCHEDULE;
     return LIP_SUCCESS;
}

int LIP_OOC_Schedule_write (LIP_Schedule s, LIP_Schedule_ooc * so)
{
    LIP_ISchedule *is;
    LIP_ISchedule_ooc * iso;
    MPI_Status status;
    int size;
    LIP_Size(&size);
    is = LIP_Schedule_convert(s);
    iso = LIP_OOC_Schedule_convert(*so);
    MPI_File_write(iso->file, is->esizes, size, MPI_INT, &status);
    MPI_File_write(iso->file, is->edispls, size, MPI_INT, &status);
    MPI_File_write(iso->file, is->elocations, 
                   is->esizes[size-1]+is->edispls[size-1], MPI_INT, &status);
    MPI_File_write(iso->file, is->isizes, size, MPI_INT, &status);
    MPI_File_write(iso->file, is->idispls, size, MPI_INT, &status);
    MPI_File_write(iso->file, &(is->data_size), 1, MPI_INT, &status);
    return LIP_SUCCESS;
}

int LIP_OOC_Schedule_reset (LIP_Schedule_ooc * s)
{
     MPI_Info info;
     LIP_ISchedule_ooc * iso;
     iso = LIP_OOC_Schedule_convert(*s);    
     MPI_Info_create(&info);
     MPI_File_set_view(iso->file, 0, MPI_INT, MPI_INT, "native", info);
     return LIP_SUCCESS;
}

int LIP_OOC_Schedule_read (LIP_Schedule s, LIP_Schedule_ooc * so)
{
    char*pname="LIP_Schedule_read";
    LIP_ISchedule *is;
    MPI_Status status;
    int size;
    LIP_ISchedule_ooc * iso;
    iso = LIP_OOC_Schedule_convert(*so);    
    LIP_Size(&size);
    is = LIP_Schedule_convert(s);
    MPI_File_read(iso->file, is->esizes, size, MPI_INT, &status);
    MPI_File_read(iso->file, is->edispls, size, MPI_INT, &status);
    is->elocations = (int *) realloc(is->elocations, sizeof (int)* 
                                      (is->esizes[size-1]+is->edispls[size-1]));
    if (NULL == is->elocations && (is->esizes[size-1]+is->edispls[size-1])!=0 )
    {
          LIP_Error_ret(pname,"for reallocating tables",LIP_OUT_OF_MEMORY);
          return LIP_OUT_OF_MEMORY;
    }
    MPI_File_read(iso->file, is->elocations, 
                  is->esizes[size-1]+is->edispls[size-1], MPI_INT, &status);
    MPI_File_read(iso->file, is->isizes, size, MPI_INT, &status);
    MPI_File_read(iso->file, is->idispls, size, MPI_INT, &status);
    MPI_File_read(iso->file, &(is->data_size), 1, MPI_INT, &status);
    return LIP_SUCCESS;
}

int LIP_OOC_Schedule_free (LIP_Schedule_ooc * s)
{
     LIP_ISchedule_ooc * iso;
     iso = LIP_OOC_Schedule_convert(*s);    
     MPI_File_close(&iso->file);
     MPI_File_delete(iso->fname, MPI_INFO_NULL);
     free(iso->fname);
     ooc_schedule_array[*s-1]=NULL;
     *s = LIP_SCHEDULE_NULL;  
     return LIP_SUCCESS;
}

LIP_ISchedule_ooc *
LIP_OOC_Schedule_convert(LIP_Schedule_ooc s)
{
  return LIP_SCHEDULE_NULL == s ? NULL : ooc_schedule_array[s - 1];
}
