/*LIP library 
setup.c - contains functions for setup environment for LIP*/

#include <lipimpl.h>
#include <debug.h>
LIP_Env LIP_env;

/*This function must be called before any other LIP function*/ 
int
LIP_Setup(int *ranks, int size)
{
  int flag;
  char*pname="LIP_Setup";

  /* error occurs if MPI was not initialized */
  if (MPI_Initialized( &flag ) != MPI_SUCCESS || ! flag){
    LIP_Error_ret(pname,"MPI not initialized",LIP_ERROR);
    return LIP_ERROR;
  }

  /* FIXME: use `ranks' and `size' arguments for proper LIP communicator
   * initialization */
  LIP_env.comm = MPI_COMM_WORLD;

  MPI_Comm_rank( LIP_env.comm, &LIP_env.rank );
  MPI_Comm_size( LIP_env.comm, &LIP_env.size );
  LIP_Schedule_array_init();
  LIP_OOC_Schedule_array_init();
  LIP_Datamap_array_init();
  LIP_IOBufmap_array_init();

#ifdef DEBUG
  LIP_DEBUG_create_file();
#endif  

  LIP_env.initialized = 1;

  return LIP_SUCCESS;
} /* LIP_Setup */

/*This function must be called after all*/
int
LIP_Exit(void)
{
  /* here all the data allocated by LIP are freed (through arrays of
   * handles represented by 'int') */
    LIP_Datamap_array_free();
    LIP_Schedule_array_free();
    LIP_OOC_Schedule_array_free();
    LIP_IOBufmap_array_free();

  return LIP_SUCCESS;
}
/*getting rank of a current process*/
int
LIP_Rank(int *rank)
{
  char *pname="LIP_Rank";
  if (1 != LIP_env.initialized){
    LIP_Error_ret(pname,"LIP not initialized",LIP_ERROR);
    return LIP_ERROR;
  }
  *rank = LIP_env.rank;

  return LIP_SUCCESS;
} /* LIP_Rank */

/*getting size of a LIP communicator*/
int
LIP_Size(int *size)
{
  char *pname="LIP_Size";
  if (1 != LIP_env.initialized){
    LIP_Error_ret(pname,"LIP not initialized",LIP_ERROR);
    return LIP_ERROR;
  }
  *size = LIP_env.size;

  return LIP_SUCCESS;
} /* LIP_Size */

